classdef Brep2D
    properties(GetAccess = 'public', SetAccess = 'private')
        myBrep; % See brepformat.txt
        myNumDivisionsPerArcSegment;
        myMinArcSegments;
        myPDEGeom;
        myDebug;
        mySegmentMapping;
        myArea;
        myNumBoundarySegments;
        myModelScale;
        mySegmentLengths;
        mySegmentType; % associated with thick (0) or thin (1) region
        myBoundingBox;
        myPolygon;
        myTitle; % figure title
    end
    methods(Static)
        function brep = readBrep2D(fileName)
            fid = fopen(fileName,'r');
            nPoints = fscanf(fid,' %d ',1);
            brep.vertices = fscanf(fid,' %f ',[2, nPoints]);
            nSegments = fscanf(fid,' %d ',1);
            brep.segments = fscanf(fid,' %f ',[4, nSegments]);
            fclose(fid);
        end
    end
    methods
        function obj = Brep2D(brep)
            % brep can be a brep structure or a file name
            % structure: vertices (2 X N), edges (2 X M)
            if (ischar(brep))
                obj.myBrep =  obj.readBrep2D(brep); % just read the file
            else
                vertices = brep.vertices;
                edges = brep.edges;
                if (size(vertices,1) ~=2)
                    vertices = vertices'; % transpose
                end
                if (size(edges,1) ~=2)
                    edges = edges';
                end
                obj.myBrep.vertices = vertices;
                % the segment structure has additional data
                nSegments = size(edges,2);
                segments = zeros(4,nSegments);
                segments(1,:) = 1; % all line segments, by default
                segments(2:3,:) = edges;
                
                if (isfield(brep,'arcs')) % if there are any arcs
                    arcs = brep.arcs;
                    if (size(arcs,1) ~=2)
                        arcs = arcs';
                    end
                    nArcs = size(arcs,2);
                    for i = 1:nArcs
                        edge = arcs(1,i);
                        vertex = arcs(2,i);
                        segments(1,edge) = 2;
                        segments(4,edge) = vertex;
                    end
                end
                
                if (isfield(brep,'virtual')) % if there are any virtual edge
                    virtual = brep.virtual;
                    if (size(virtual,1) ~=1)
                        virtual = virtual';
                    end
                    nVirtualEdges = size(virtual,2);
                    for i = 1:nVirtualEdges
                        edge = virtual(1,i);
                        segments(1,edge) = -1;
                    end
                end
                obj.myBrep.segments = segments;
            end
            obj.myNumBoundarySegments = size(obj.myBrep.segments,2);
            obj.myArea = obj.brepArea();
            obj.myPolygon = obj.computePolygon();
            obj.myNumDivisionsPerArcSegment = 6;
            x = obj.myBrep.vertices(1,:);
            y = obj.myBrep.vertices(2,:);
            obj.myModelScale = ((max(x)-min(x))+(max(y)-min(y)))/2;
            obj.mySegmentLengths = zeros(1,obj.myNumBoundarySegments);
            obj = obj.convertBreptoPdeGeom();
            obj.myBoundingBox = [min(x) max(x); min(y) max(y)];
            
        end
        function obj = resetBrep(obj,brep)
            obj.myBrep = brep;
            obj.myArea = obj.brepArea();
            obj = obj.convertBreptoPdeGeom(obj.myNumDivisionsPerArcSegment);
            obj.myPolygon = obj.computePolygon();
        end
        function obj =  convertBreptoPdeGeom(obj,hMax)
            % convert from brep format to Matlab's pdegeom format
            brep = obj.myBrep;
            v = brep.vertices;
            nSegments = size(brep.segments,2);
            pdeGeom = zeros(11,1);
            segmentMapping = [];
            npdeGeomSegments = 0;
            dl = 1; dr = 0;
            for i = 1:nSegments
                breptype = brep.segments(1,i);
                vs = brep.segments(2,i);
                ve = brep.segments(3,i);
                
                xs = v(1,vs);xe = v(1,ve);
                ys = v(2,vs);ye = v(2,ve);
                if (breptype == 1) % line
                    pdetype = 2;
                    npdeGeomSegments = npdeGeomSegments+1;
                    pdeGeom(:,npdeGeomSegments) = [pdetype xs xe ys ye dl dr 0 0 0 0];
                    segmentMapping(npdeGeomSegments) = i;
                    obj.mySegmentLengths(i) = sqrt( (xe-xs)^2 + (ye-ys)^2);
                elseif (breptype == 2) % arc
                    pdetype = 1;
                    vc = abs(brep.segments(4,i));
                    xc = v(1,vc); yc  = v(2,vc);
                    R = sqrt((xc-xs)^2+(yc-ys)^2);
                    % # of segments generated will depend on hmax
                    d = sqrt((xe-xs)^2+(ye-ys)^2);
                    theta = real(2*asin(d/2/R)); % due to floating point errors, sometimes d > 2*R resulitng in complex values
                    arcLength = R*theta;
                    obj.mySegmentLengths(i) =arcLength;
                    if (nargin == 2) % hMax given
                        numArcSegments = max(ceil(0.25*arcLength/hMax),2*theta/(pi/2));
                    else
                        numArcSegments = obj.myNumDivisionsPerArcSegment;
                    end
                    vecStart = [xs-xc ys-yc];
                    vecStart = vecStart/norm(vecStart);
                    vecEnd = [xe-xc ye-yc];
                    vecEnd = vecEnd/norm(vecEnd);
                    thetaStart = atan2(vecStart(2),vecStart(1));
                    thetaEnd = atan2(vecEnd(2),vecEnd(1));
                    if (brep.segments(4,i) > 0) % cw
                        if (thetaEnd > thetaStart) % not allowed
                            thetaEnd = thetaEnd - 2*pi;
                        end
                    else % ccw
                        if (thetaEnd < thetaStart) % not allowed
                            thetaEnd = thetaEnd + 2*pi;
                        end
                    end
                    for j = 1:numArcSegments
                        npdeGeomSegments = npdeGeomSegments+1;
                        segmentMapping(npdeGeomSegments) = i; %#ok<*AGROW>
                        theta0 = thetaStart+(j-1)*(thetaEnd-thetaStart)/numArcSegments;
                        theta1= thetaStart+j*(thetaEnd-thetaStart)/numArcSegments;
                        xs = xc + R*cos(theta0); xe = xc + R*cos(theta1);
                        ys = yc + R*sin(theta0); ye = yc + R*sin(theta1);
                        if (brep.segments(4,i) > 0) % clockwise
                            pdeGeom(:,npdeGeomSegments) = [pdetype xe xs ye ys dr dl xc yc R R];
                        else % anti-clockwise
                            pdeGeom(:,npdeGeomSegments) = [pdetype xs xe ys ye dl dr xc yc R R];
                        end
                    end
                elseif (breptype == 3) % circle ... start and end vertices are same
                    pdetype = 1; % arc
                    vc = abs(brep.segments(4,i));
                    xc = v(1,vc); yc  = v(2,vc);
                    R = sqrt((xc-xs)^2+(yc-ys)^2);
                    
                    % # of segments generated will depend on hmax
                    arcLength = R*2*pi;
                    if (nargin == 2) % hMax given
                        numArcSegments = max(ceil(0.25*arcLength/hMax),2*theta/(pi/2));
                    else
                        
                        numArcSegments = obj.myNumDivisionsPerArcSegment;
                    end
                    for j = 1:numArcSegments
                        npdeGeomSegments = npdeGeomSegments+1;
                        segmentMapping(npdeGeomSegments) = i;
                        thetaStart = (j-1)*2*pi/numArcSegments;
                        thetaEnd = j*2*pi/numArcSegments;
                        
                        xs = xc + R*cos(thetaStart); xe = xc + R*cos(thetaEnd);
                        ys = yc + R*sin(thetaStart); ye = yc + R*sin(thetaEnd);
                        pdeGeom(:,npdeGeomSegments) = [pdetype xs xe ys ye dl dr xc yc R R];
                    end
                end
            end
            obj.myPDEGeom = pdeGeom;
            obj.mySegmentMapping = segmentMapping;
        end
         function obj = saveBrepFile(obj,fileName)
            brep = obj.myBrep;
            v = brep.vertices;
            seg = brep.segments;
            fid = fopen(fileName,'w');
            fprintf(fid,'%d \n',size(v,2));
            fprintf(fid,'%f %f \n',v);
            fprintf(fid,'%d \n',size(seg,2));
            fprintf(fid,'%d %d %d %d \n',seg);
            fclose(fid);
        end
        function pts = sampleBoundary(obj,h)
            brep = obj.myBrep;
            v = brep.vertices;
            nSegments = size(brep.segments,2);
            pts = [];
            for i = 1:nSegments
                breptype = brep.segments(1,i);
                vs = brep.segments(2,i);
                ve = brep.segments(3,i);
                
                xs = v(1,vs);xe = v(1,ve);
                ys = v(2,vs);ye = v(2,ve);
                if (breptype == 1) % line
                    L = norm([xe-xs ye-ys]);
                    n = ceil(L/h) + 1;
                    if (rem(n,2) == 0)
                        n = n + 1; % make sure number of nodes is odd
                    end
                    eta = (0:n-2)/(n-1);
                    p = [xs + eta*(xe-xs); ys + eta*(ye-ys)];
                    pts = [pts p];
                end
            end
        end
        function sym = isGeometrySymmetricAboutX(obj)
            % only checks for a necessary condition of vertex symmetry
            sym = 0;
            v = obj.myBrep.vertices;
            nVertices = size(v,2);
            v(2,:) = v(2,:)- mean(v(2,:)); % shift all vertices
            for i = 1:nVertices
                x0 = v(1,i);
                y0 = v(2,i);
                if (abs(y0) < 100*eps), continue; end % point lies on symmetry line
                found = 0;
                for j = 1:nVertices
                    % check if there exists another point such that
                    % x1 = x0, and y1 = -y0;
                    x1 = v(1,j);
                    y1 = v(2,j);
                    if ((abs(y0 + y1) < 100*eps) && (abs(x0-x1) < 100*eps)), found = 1; break;end
                end
                if (~found), return;end
            end
            sym = 1;
        end
        function sym = isGeometrySymmetricAboutY(obj)
            % only checks for a necessary condition of vertex symmetry
            sym = 0;
            v = obj.myBrep.vertices;
            nVertices = size(v,2);
            v(1,:) = v(1,:)- mean(v(1,:)); % shift all vertices
            for i = 1:nVertices
                x0 = v(1,i);
                y0 = v(2,i);
                if (abs(x0) < 100*eps), continue; end % point lies on symmetry line
                found = 0;
                for j = 1:nVertices
                    % check if there exists another point such that
                    % x1 = -x0, and y1 = y0;
                    x1 = v(1,j);
                    y1 = v(2,j);
                    if ((abs(y0 - y1) < 100*eps) && (abs(x0+x1) < 100*eps)), found = 1; break;end
                end
                if (~found), return;end
            end
            sym = 1;
        end
        function polygon = computePolygon(obj)
            % Convert the brep into Matlab polygon structure, useful for
            % point membership. Arcs are approximated by  single straight lines!!
            brep = obj.myBrep;
            v = brep.vertices;
            nSegments = size(brep.segments,2);
            xv(1,1) = v(1,1);
            yv(1,1) = v(2,1);
            counter = 2;
            for i = 1:nSegments
                if(brep.segments(1,i)== -1)
                    xv(counter) = NaN;
                    yv(counter) = NaN;
                    counter = counter + 1;
                    node = brep.segments(3,i);
                    xv(counter) = v(1,node);
                    yv(counter) = v(2,node);
                    counter = counter+1;
                elseif (brep.segments(1,i)== 1)
                    node = brep.segments(3,i);
                    xv(counter) = v(1,node);
                    yv(counter) = v(2,node);
                    counter = counter+1;
                elseif (brep.segments(1,i)== 2)
                    node1 = brep.segments(2,i);
                    node2 = brep.segments(3,i);
                    center = abs(brep.segments(4,i));
                    dir = -sign(brep.segments(4,i));
                    nSections = 20;
                    [X,Y] = obj.PtsOnArc(node1,node2,center,nSections,dir);
                    xv(counter:counter+nSections-1) = X;
                    yv(counter:counter+nSections-1) = Y;
                    counter = counter + nSections;
                end
                
            end
            polygon0 = [xv;yv];
            polygon = zeros(size(polygon0));
            k = 0;
            counter = 0;
            % outside boundary
            for i = 1:size(polygon,2)
                if (isnan(polygon0(:,i)))
                    k = k+1;
                    continue;
                end
                if (mod(k,2)==0)
                    counter = counter + 1;
                    polygon(:,counter) = polygon0(:,i);
                end
            end
            % holes
            k = 0;
            for i = 1:size(polygon,2)
                if (isnan(polygon0(:,i)))
                    k = k+1;
                    counter = counter + 1;
                    polygon(:,counter) = polygon0(:,i);
                    continue;
                end
                if (mod(k,2)==1)
                    counter = counter + 1;
                    polygon(:,counter) = polygon0(:,i);
                end
            end
            % fix
            if (isnan(polygon(:,size(polygon,2))))
                polygon(:,size(polygon,2)) = [];
            end
        end
        function in = isInside(obj,pts)
            % uses inpolygon to compute point membership
            polygon = obj.computePolygon();
            in = inpolygon(pts(1,:),pts(2,:),polygon(1,:),polygon(2,:));
        end
        function [dMin,closestPts,closestSegs] = distOfPointsToBrep(obj,pts,segmentsToCheck)
            brep = obj.myBrep;
            v = brep.vertices;
            nSegments = size(brep.segments,2);
            if (nargin == 2)
                segmentsToCheck = 1:nSegments;
            end
            nPts = size(pts,2);
            dMin = 1e12*ones(1,nPts);
            closestPts = pts; % default
            closestSegs = zeros(1,nPts);
            for seg = segmentsToCheck
                breptype = brep.segments(1,seg);
                if (breptype == 1) % line
                    vs = brep.segments(2,seg);
                    ve = brep.segments(3,seg);
                    [dSeg,closestPtsSeg] = obj.distOfPointsToLineSegment(pts,v(:,vs),v(:,ve));
                elseif (breptype == 2) % arc
                    [dSeg,closestPtsSeg] = obj.distOfPointsToArcSegment(pts,seg);
                end
                for n = 1:nPts
                    if (dSeg(n) < dMin(n))
                        dMin(n) = dSeg(n);
                        closestPts(:,n) = closestPtsSeg(:,n);
                        closestSegs(n) = seg;
                    end
                end
            end
        end
        function [X,Y] = PtsOnArc(obj,node1,node2,center,nSections,dir)
            v = obj.myBrep.vertices;
            x1 = v(1,node1);
            y1 = v(2,node1);
            x2 = v(1,node2);
            y2 = v(2,node2);
            xc = v(1,center);
            yc = v(2,center);
            % find angle
            r1 = [x1-xc;y1-yc];
            r2 = [x2-xc;y2-yc];
            radius = norm(r1);
            xaxis = [1;0];
            Startangle = acos(dot(r1,xaxis)/radius);
            Endangle = Startangle + acos(dot(r1,r2)/radius^2);
            delta = dir*abs(Endangle-Startangle)/nSections;
            cp = cross([xaxis;1],[r1;1]);
            if (cp(3) < 0)
                Startangle = Startangle + pi;
            end
            Angle = Startangle;
            X(1) = x1;
            Y(1) = y1;
            for i = 2:nSections+1
                Angle = Angle+delta;
                X(i) = xc+ radius*cos(Angle);
                Y(i) = yc + radius*sin(Angle);
            end
            X = X(2:end);
            Y = Y(2:end);
        end
        function [d,closestPts] = distOfPointsToLineSegment(~,pts,lineStart,lineEnd)
            % vectorized operation
            ABS_TOL = 1e-10;
            L = norm(lineStart-lineEnd);
            closestPts = pts;
            if (L < ABS_TOL)
                d = norm(pts-lineStart);
                closestPts(1,:) = lineStart(1);
                closestPts(2,:) = lineStart(2);
                return;
            end
            lineTangent = (lineEnd-lineStart)/L;
            v1(1,:) = pts(1,:) - lineStart(1);
            v1(2,:) = pts(2,:) - lineStart(2);
            v2(1,:) = pts(1,:) - lineEnd(1);
            v2(2,:) = pts(2,:) - lineEnd(2);
            numer = (lineEnd-lineStart)'*v1;
            u = numer/L^2;
            distToStart = norm(v1);
            distToEnd= norm(v2);
            distToLineSeg = abs((-v1(1,:)*lineTangent(2) + v1(2,:)*lineTangent(1)));
            d = (u < 0).*distToStart + (u >1).*distToEnd + (u>=0).*(u<= 1).*distToLineSeg;
            closestPts(1,:) = (u < 0).*lineStart(1) + (u >1).*lineEnd(1) + ...
                (u>=0).*(u<= 1).*((1-u).*lineStart(1) + u.*lineEnd(1));
            closestPts(2,:) = (u < 0).*lineStart(2) + (u >1).*lineEnd(2) + ...
                (u>=0).*(u<= 1).*((1-u).*lineStart(2) + u.*lineEnd(2));
        end
        function [d,closestPts] = distOfPointsToArcSegment(obj,pts,seg)
            brep = obj.myBrep;
            vc = abs(brep.segments(4,seg));
            v = brep.vertices;
            xc = v(1,vc); yc  = v(2,vc);
            vs = brep.segments(2,seg);
            ve = brep.segments(3,seg);
            xs = v(1,vs); ys = v(2,vs);
            xe = v(1,ve); ye = v(2,ve);
            R = sqrt((xc-xs)^2+(yc-ys)^2);
            vecStart = [xs-xc ys-yc];
            vecStart = vecStart/norm(vecStart);
            vecEnd = [xe-xc ye-yc];
            vecEnd = vecEnd/norm(vecEnd);
            thetaStart = atan2(vecStart(2),vecStart(1));
            thetaEnd = atan2(vecEnd(2),vecEnd(1));
            if (brep.segments(4,seg) > 0) % cw
                if (thetaEnd > thetaStart) % not allowed
                    thetaEnd = thetaEnd - 2*pi;
                end
            else % ccw
                if (thetaEnd < thetaStart) % not allowed
                    thetaEnd = thetaEnd + 2*pi;
                end
            end
            % sample the arc
            N  = 100;
            theta = thetaStart + (0:N)*(thetaEnd-thetaStart)/N;
            arcPts = zeros(2,N+1);
            arcPts(1,:) = xc + R*cos(theta);
            arcPts(2,:)  = yc + R*sin(theta);
            % find closest pt
            closestPts = pts;
            d = zeros(1,size(pts,2));
            for i = 1:size(pts,2)
                dArc = sqrt( (pts(1,i) - arcPts(1,:)).^2 +  (pts(2,i) - arcPts(2,:)).^2);
                [d(i),index] = min(dArc);
                closestPts(:,i) = arcPts(:,index);
            end
            
        end
        function adjustFigScale(~)
            axis tight;
            v = axis;
            xRange = v(2)-v(1);
            yRange = v(4)-v(3);
            deltaX = 0.1*xRange;
            deltaY = 0.1*yRange;
            v = [v(1)-deltaX v(2)+deltaX  v(3)-deltaY v(4)+deltaY];
            axis(v);
        end
        function obj = setTitle(obj,title)
            obj.myTitle = title;
            obj.adjustFigScale();
        end
        function plotGeometry(obj)
            pdeGeom = obj.myPDEGeom;
            pdegplot(pdeGeom);
            obj.adjustFigScale();
            axis equal;
        end
        function plotGeometryWithLabels(obj)
            pdeGeom = obj.myPDEGeom;
            segmentMapping = obj.mySegmentMapping;
            pdegplot(pdeGeom);
            obj.adjustFigScale();
            segments = 1:size(pdeGeom,2);
            for seg = segments
                brepSegment = segmentMapping(seg);
                xmid = mean(pdeGeom(2:3,seg));
                ymid = mean(pdeGeom(4:5,seg));
                text(xmid,ymid,num2str(brepSegment),'fontsize',16);
            end
            axis equal;
            hold on;
            p = obj.myBrep.vertices;
            for i = 1:size(p,2)
                plot(p(1,i),p(2,i),'b*');
                text(p(1,i),p(2,i),num2str(i),'Color','red','fontsize',16);
            end
            obj.adjustFigScale();
        end
        
        function area = areaUnderLineSegment(~,startPt,endPt)
            base = (startPt(1)-endPt(1));
            a1 = base*startPt(2);
            a2 = 0.5*(base)*(endPt(2)-startPt(2));
            area = a1 + a2;
        end
        function area = areaUnderArcSegment(obj,startPt,endPt,centerPt)
            a1 = obj.areaUnderLineSegment(centerPt,endPt);
            a2 = obj.areaUnderLineSegment(centerPt,startPt);
            R = norm(endPt-centerPt);
            d = norm(endPt-startPt);
            theta = real(2*asin(d/2/R));
            a3 = theta*R^2/2;
            area = -a1 + a2 +a3;
        end
        
        
        function area = brepArea(obj)
            brep = obj.myBrep;
            v = brep.vertices;
            nSegments = size(brep.segments,2);
            area = 0;
            for i = 1:nSegments
                breptype = brep.segments(1,i);
                vs = brep.segments(2,i);
                ve = brep.segments(3,i);
                if (breptype == 1) % line
                    area = area + obj.areaUnderLineSegment(v(:,vs),v(:,ve));
                elseif (breptype == 2) % arc
                    vc = abs(brep.segments(4,i));
                    a = obj.areaUnderArcSegment(v(:,vs),v(:,ve),v(:,vc));
                    if (brep.segments(4,i) < 0)
                        area = area + a;
                    else
                        area = area -a;
                    end
                elseif (breptype == 3) % circle
                    vc = abs(brep.segments(4,i));
                    radius = norm(v(:,vs)-v(:,vc));
                    if (i == 1)
                        area = area + pi*radius^2; %external boundary
                    else
                        area = area - pi*radius^2; %internal hole
                    end
                elseif (breptype == -1) % internal segment
                    % do nothing
                else
                    disp('Unknown breptype type');
                end
            end
        end
        
        function obj = setDebugOn(obj)
            obj.myDebug = 1;
        end
        function obj = setDebugOff(obj)
            obj.myDebug = 0;
        end
        function drawArrow(~,start,stop,clr,thickness)
            if (nargin == 3)
                clr = 'r';
                thickness = 1.0;
            end
            for i = 1:size(start,1)
                x0 = start(i,1);
                y0 =start(i,2);
                x1 = stop(i,1);
                y1 = stop(i,2);
                
                h = plot([x0 x1],[y0, y1],clr);
                set(h,'linewidth',thickness);
                hold on;
                p = stop(i,:)-start(i,:);
                alpha = 0.1;  % Size of arrow head relative to the length of the vector
                beta = 0.35;  % Width of the base of the arrow head relative to the length
                hu = [x1-alpha*(p(1)+beta*(p(2)+eps)); x1; x1-alpha*(p(1)-beta*(p(2)+eps))];
                hv = [y1-alpha*(p(2)-beta*(p(1)+eps)); y1; y1-alpha*(p(2)+beta*(p(1)+eps))];
                h = plot(hu(:),hv(:),clr);  % Plot arrow head
                set(h,'linewidth',thickness);
            end
        end
        
    end
end
