classdef solidLab
    % An interface between SolidWorks and Matlab
    properties
        myHome;
        myVersion;
        mySolidWorksConnection;
        mySimulationConnection;
    end
    methods
        function obj = solidLab
            obj.myVersion = 2023.02;
            obj.mySolidWorksConnection = 0;
            obj.mySimulationConnection = 0;
            if (isempty(getenv('SOLIDLAB_HOME')))
                disp('Please set SOLIDLAB_HOME');
                return;
            end
            obj.myHome = getenv('SOLIDLAB_HOME');
            obj.myHome = ['"'  obj.myHome  '"']; % need to add quotes to address issues with spaces
            addpath(obj.myHome);
            
            if (system([obj.myHome '\dist\pyworksinterface 1']))
                disp(['SolidLab Home: ' obj.myHome]);
                disp(['SolidLab Version: ' num2str(obj.myVersion)]);
                disp('Connection to SolidWorks established.');
                obj.mySolidWorksConnection = 1;
            else 
                disp('Connection to SolidWorks could NOT be established.');
                disp('Make sure SolidWorks is running with a part open.');
                disp('If you have more than one version of SolidWorks installed, use the latest version.');
                
            end
            if (exist('solidLabTemp.txt','file'))
                delete('solidLabTemp.txt'); % start afresh
            end
            path(path,genpath(cd));
        end
        function obj = openSolidWorksPart(obj,filename,dir)
            %if directory not provided, assume current dir
            if(nargin == 2)
                dir ='.';
            end
            fn = strcat(dir,'\',filename);
            if(~exist(fn,'file'))
                message = strcat('openSolidWorksPart: The file "',filename,'" does not exist in the specified directory.');
                disp(message);
                return;
            end
            filename = strcat(strrep(dir,'\','\\'), '\\', filename);
            disp(filename);
            if (~system([obj.myHome '\dist\pyworksinterface 13 ' filename]))
                disp('openSolidWorksPart: Failed');
                obj.mySolidWorksConnection= 0;
                return;
            end
            obj.mySolidWorksConnection = 1;
        end
        function [volume,area,cg,inertia] = getMassProperties(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 2']))
                disp('getMassProperties: Failed');
                volume=0;area=0;cg=0;inertia=0;
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            res = fscanf(f,'[%e,%e,%e,%e,%e,%e,%e,%e,%e,%e,%e]');
            fclose(f);
            volume = res(1);
            area = res(2);
            cg = res(3:5);
            I = res(6:11);
            inertia = [I(1) I(4) I(5); I(4) I(2) I(6); I(5) I(6) I(3)];
        end
        function [edgeCount] = getEdgeCount(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 29']))
                disp('getEdgeCount: Failed');
                edgeCount= 0;
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            edgeCount = fscanf(f,'%d');
            fclose(f);
        end
        function [faceCount] = getFaceCount(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 31']))
                disp('getFaceCount: Failed');
                faceCount= 0;
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            faceCount = fscanf(f,'%d');
            fclose(f);
        end
        function closeAllDocuments(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 30']))
                disp('closeAllDocuments: Failed');
                return;
            end
        end
        function [dimNames,dimValues] = getDimensions(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 3']))
                disp('getDimensions: Failed');
                dimNames = [];
                dimValues = [];
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            nDim = fscanf(f,'%d',1);
            dimNames = cell(1,nDim);
            dimValues = zeros(1,nDim);
            for i = 1:nDim
                dimNames{i} = fscanf(f,'%s',1);
                dimValues(i) = fscanf(f,'%f',1);
            end
            fclose(f);
        end
        
        function  modifyDimensions(obj,dimNames,dimValues)
            if (~iscell(dimNames))
                dimNames = {dimNames};
            end
            nDim = numel(dimNames);
            if (nDim == 0)
                disp('Zero dimensions provided')
                return;
            end
            % write the dimensions to a file
            f = fopen('solidLabTemp.txt','w');
            for i = 1:nDim
                fprintf(f,'%s %e\n',dimNames{i},dimValues(i));
            end
            fclose(f);
            % read and modify solidworks
            if (~system([obj.myHome '\dist\pyworksinterface 4']))
                disp('modifyDimensions: Failed');
                return;
            end
        end

         function  suppress(obj,partName)
            % write the part name to a file
            if (~system([obj.myHome '\dist\pyworksinterface 10 ' partName]))
                disp('suppress: Could not establish connection to SolidWorks');
                return;
            end
        end
        function unsuppress(obj,partName)
            if (~system([obj.myHome '\dist\pyworksinterface 11 ' partName]))
                disp('unsuppress: Could not establish connection to SolidWorks');
                return;
            end
        end
        function saveSolidWorksPart(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 16']))
                disp('saveSolidWorksPart: Could not establish connection to SolidWorks'); 
                return;
            end
        end
        function [featureNames,featureTypes, extrudeDepthForward, extrudeDepthReverse] = getAllFeatures(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 32']))
                disp('getAllFeatuures: Could not establish connection to SolidWorks'); 
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            nFeatures = fscanf(f,'%d',1);
            featureNames = cell(1,nFeatures);
            featureTypes = cell(1,nFeatures);
            extrudeDepthForward = zeros(1,nFeatures);
            extrudeDepthReverse = zeros(1,nFeatures);
            for i = 1:nFeatures
                featureNames{i} = fscanf(f,'%s',1);
                featureTypes{i} = fscanf(f,'%s',1);
                extrudeDepthForward(i) = fscanf(f,'%f',1);
                extrudeDepthReverse(i) = fscanf(f,'%f',1);
            end
            fclose(f);
        end
        function [brep,planeNormal,planePt] = convertSketch2Brep(obj,sketchName)
            if (~system([obj.myHome '\dist\pyworksinterface 18 ' sketchName]))
                disp('convertSketch2Brep: Failed');
                return;
            end
            obj.mySolidWorksConnection= 1;
            fid = fopen('sketchBrep.txt','r');
            nPoints = fscanf(fid,' %d ',1);
            xyz = fscanf(fid,' %f ',[3, nPoints]);
            planeNormal = [0 0 1]'; % assume xyz plane
            planePt = [0 0 0]'; % assume origin is contained in sketch plane
            brep.vertices = xyz(1:2,:); % skip for now, we need to project
            brep.vertices = unique(brep.vertices','rows');
            nPoints = numel(brep.vertices(:,1));
            nLines = fscanf(fid,' %d ',1);
            lineCoord = fscanf(fid, ' %f ',[6, nLines]);
            nArcs = fscanf(fid,' %d ',1);
            arcCoord = fscanf(fid, ' %f ',[10,nArcs]);
            [nSegments,brep.segments] = obj.getBrepSegments(brep.vertices,nLines,lineCoord,nArcs,arcCoord);
            brepvertices = brep.vertices;
            [brep_Segments,brep_Vertices] = obj.FixConnectivity(brepvertices,brep.segments,nSegments);
            fclose(fid);
            brep.vertices = brep_Vertices';
            brep.segments = brep_Segments';
            [brep.vertices,brep.segments] = obj.clearRepeatedVertices(brep.vertices,brep.segments,nPoints);
            nVerticesNew = numel(brep.vertices(1,:));
            nSegmentsNew = numel(brep.segments(1,:));
            fid2 = fopen('sketchBrep2D.txt','w');
            fprintf(fid2,'%d\r\n',nVerticesNew);
            fprintf(fid2,'%f %f\r\n',brep.vertices);
            fprintf(fid2,'%d\r\n',nSegmentsNew);
            fprintf(fid2,'%i %i %i %i\r\n',brep.segments);
            fclose(fid2);
        end
        
        function createMesh(obj,relativeSize,studyName)
            quality = 1;
            if (nargin == 1)
                relativeSize = 1.0;
                studyName = '';
            elseif (nargin == 2)
                studyName = '';
            end
            if (~system([obj.myHome '\dist\pyworksinterface 5 ' num2str(quality) ' ' num2str(relativeSize) ' ' studyName]))
                disp('createMesh: SolidWorks Simulation may not be active; activate and run FEA in SolidWorks ');
                return;
            end
        end
        
        function [nodes,elems] = getMesh(obj,studyName)
            if (nargin == 1)
                 studyName = '';
            end
            if (~system([obj.myHome '\dist\pyworksinterface 6  ' studyName]))
                disp('getMesh: SolidWorks Simulation may not be active; activate and run FEA in SolidWorks ');
                nodes = [];
                elems = [];
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            nNodes = fscanf(f,'%d');% read nNodes
            nodes = zeros(nNodes,3);
            fscanf(f,'(');
            for i=1:nNodes
                result = fscanf(f,'%d, %f, %f, %f,',4);
                nodes(result(1),:) = result(2:4);
            end
            fscanf(f,')');
            
            nElems = fscanf(f,'%d');% read nElems
            elems = zeros(nElems,4);
            fscanf(f,'(');
            for i=1:nElems
                result = fscanf(f,'%d, %d, %d, %d, %d, ',5);
                elems(result(1),:) = result(2:5);
                fscanf(f,'%f,',11);
            end
            fscanf(f,')');
            fclose(f);
        end
 
        function  dumpDraftMesh(obj,fileName,studyName)
            % usage: s.dumpDraftMesh('myMesh.tetmesh');
            if (nargin == 2)
                studyName = '';
            end
            %obj.createMesh(0);
            [nodes,elems] = obj.getMesh(studyName);
            nNodes = size(nodes,1);
            nElems = size(elems,1);
            if (nNodes == 0)
                disp('dumpDraftMesh: SolidWorks Simulation may not be active; activate and run FEA in SolidWorks ');
                return;
            end
              % create the edge graph from tet mesh
            connectivity = zeros(nNodes,nNodes);% bad idea but...
            for elem = 1:nElems
                elemNodes = elems(elem,:);
                connectivity(elemNodes,elemNodes) = 1;
            end
            nEdges =  0;
            for node = 1:nNodes
                connectivity(node,1:node) = 0;
                nEdges = nEdges + sum(connectivity(node,:));
            end
            edges = zeros(nEdges,2);
            count = 1;
            for node = 1:nNodes
                otherNodes = find(connectivity(node,:) == 1);
                for j = 1:length(otherNodes)
                    edges(count,1) = node;
                    edges(count,2) = otherNodes(j);
                    count = count +1;
                end
            end
            f = fopen(fileName,'w');
            fprintf(f,'%d \n',nNodes);
            fprintf(f,'%e %e %e \n',nodes');
            fprintf(f,'%d \n',nElems);
            fprintf(f,'%d %d %d %d\n',elems');
            fprintf(f,'%d \n',nEdges);
            fprintf(f,'%d %d \n',edges');
            fclose(f);
        end
        function  solveStudy(obj,studyName)
            if (nargin == 1)
                studyName = '';
            end
            if (~system([obj.myHome '\dist\pyworksinterface 7 ' studyName]))
                disp('solveStudy: SolidWorks Simulation may not be active; activate and run FEA in SolidWorks ');  
                return;
            end
        end
        function [vonMises,maxDisp] = getResults(obj,studyName)
            if (nargin == 1)
                studyName = '';
            end
            if (~system([obj.myHome '\dist\pyworksinterface 8 ' studyName]))
                disp('getResults: SolidWorks Simulation may not be active; activate and run FEA in SolidWorks ');
                vonMises = 0;
                maxDisp = 0;
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            vonMises = fscanf(f,'%e',1);% read stress
            maxDisp = fscanf(f,'%e',1);% read displacement
            fclose(f);
        end
       
        function [nodeStress] = getVonMisesStressAtAllNodes(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 27 ']))
                disp('getVonMisesStressAtAllNodes: Failed');
                disp('getVonMisesStressAtAllNodes: SolidWorks Simulation may not be active; activate and run FEA in SolidWorks ');
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            nodeStress = zeros(500000,1);% preallocate memory
            fscanf(f,'(');
            count = 1;
            while (1)
                result = fscanf(f,'%d, %f, %f, %f,%f, %f, %f,%f,%f,%f,%f,%f',12);
                nodeStress(count) = result(12);
                c = fscanf(f,'%c',1);
                if (c == ')')
                    break;
                else
                    count = count + 1;
                end
            end
            fscanf(f,')');
            fclose(f);
            nodeStress = nodeStress(1:count,1);% trim
        end      
        function [elemStress] = getVonMisesStressAtAllElements(obj)
            if (~system([obj.myHome '\dist\pyworksinterface 28 ']))
                disp('getVonMisesStressAtAllElements: Failed');
                disp('getVonMisesStressAtAllElements: SolidWorks Simulation may not be active; activate and run FEA in SolidWorks ');
                return;
            end
            f = fopen('solidLabTemp.txt','r');
            elemStress = zeros(500000,1);% preallocate memory
            fscanf(f,'(');
            count = 1;
            while (1)
                result = fscanf(f,'%d, %f, %f, %f,%f, %f, %f,%f,%f,%f,%f,%f',12);
                elemStress(count) = result(12);
                c = fscanf(f,'%c',1);
                if (c == ')')
                    break;
                else
                    count = count + 1;
                end
            end
            fscanf(f,')');
            fclose(f);
            elemStress = elemStress(1:count,1);% trim
        end
        function [pNormElemStress] = getElementPNormStress(obj,pNorm)
            if (nargin == 1)
                pNorm = 6;
            end
            [elemStress] = obj.getVonMisesStressAtAllElements();
            pNormElemStress = (sum(elemStress.^pNorm)).^(1/pNorm);
        end
        function [pNormNodeStress] = getNodePNormStress(obj,pNorm)
            if (nargin == 1)
                pNorm = 6;
            end
            [nodeStress] = obj.getVonMisesStressAtAllNodes();
            pNormNodeStress = (sum(nodeStress.^pNorm)).^(1/pNorm);
        end
    end
end
