classdef spring2dPlotter
    properties(GetAccess = 'public', SetAccess = 'public')
        % public read access, but private write access.
        myNodeLocations; % (x,y) node locations,(2, N), N is number of nodes
        myConnectivity; % (startnode, endnode) (2, M), M is number of spring elements
        myNumNodes; % number of Nodes
        myNumSpringElements; % number of spring elements
        myL0; % Initial length of each spring element
        myTheta0; % Initial orientation of each spring element
        myDeformations; % deformation of each element
        myDeformationMax; % max deformation among all members
        myUV;
    end
    methods
        function obj = spring2dPlotter(nodeXY,connectivity)
            % constructor for a spring system
            % nodeXY: 2 x N, N is the number of nodes
            % connectivity: 2 x M, M is the number of spring elements
            if (size(nodeXY,1) ~= 2)
                disp('Error: nodeXY must be of size 2 X nNodes');
                obj  = [];
                return;
            end
            if (size(connectivity,1) ~= 2)
                disp('Error: connectivity must be of size 2 X nSprings');
                obj  = [];
                return;
            end
            obj.myNodeLocations = nodeXY;
            obj.myUV  = 0*nodeXY;
            obj.myConnectivity = connectivity;
            obj.myNumNodes = size(nodeXY,2);
            obj.myNumSpringElements = size(connectivity,2);
            % Find the length and orientation of each element
            C = obj.myConnectivity;
            startPts = obj.myNodeLocations(:,C(1,:));
            endPts = obj.myNodeLocations(:,C(2,:));
            obj.myL0 = sqrt((endPts(1,:)-startPts(1,:)).^2 + (endPts(2,:)-startPts(2,:)).^2);
            obj.myDeformationMax = 0;
            obj.myTheta0 = atan2(endPts(2,:)-startPts(2,:),endPts(1,:)-startPts(1,:));
        end
        function [x,y] = springTemplate(~)
            N = 20;
            x = zeros(2*N+3,1);
            y = zeros(2*N+3,1);
            x(1) = 0; y(1) = 0;
            x(2) = 0.05; y(2) = 0;
            delta = 0.9/(2*N);
            x0 = 0.05 - delta;
            x(3:2:2*N+2) = x0 + delta*(2:2:2*N+1);
            x(4:2:2*N+3) = x(3:2:2*N+2) + delta;
            x(2*N+2) = 0.95;
            x(2*N+3) = 1;
            y(3:2:2*N+2) = 0.025;
            y(4:2:2*N+3) = -0.025;
        end
        function plot(obj)
            N = obj.myNumNodes; % number of nodes
            P = obj.myNodeLocations; % extract undeformed location
            % Length of each spring
            L0 = obj.myL0;
            LScale = max(L0);
            [xTemplate,yTemplate] = obj.springTemplate();
            ct = cos(obj.myTheta0);
            st = sin(obj.myTheta0);
            xy = obj.myNodeLocations;
            for s = 1:obj.myNumSpringElements
                xStretch = xTemplate*L0(s);
                yStretch = yTemplate;
                xRotate = xStretch*ct(s) - yStretch*st(s);
                yRotate = xStretch*st(s) + yStretch*ct(s);
                xSpring = xRotate + xy(1,obj.myConnectivity(1,s));
                ySpring = yRotate + xy(2,obj.myConnectivity(1,s));
                plot(xSpring,ySpring,'b','linewidth',1.0);hold on;
                xText = 0.4*xSpring(1)+0.6*xSpring(end) +0.05*LScale; % to avoid overlapping texts
                yText = 0.4*ySpring(1)+0.6*ySpring(end);
                text(xText,yText,num2str(s),'FontAngle','italic','FontSize',14);
            end
            for j = 1:N % for all nodes
                x0 = P(1,j); y0 = P(2,j);
                h =plot(x0,y0,'ro'); set(h,'LineWidth',0.25);
                text(x0+0.05*LScale,y0,num2str(j),'Color',[1 0 0],'FontSize',14);      
            end
            axis equal;axis fill;
        end

        function obj = setUVAtNode(obj,node,uv)
            obj.myUV(:,node) = uv(:);
        end
        function plotDeformed(obj)
            uv = obj.myUV;
            u = uv(1,:);
            v = uv(2,:);
            deformation = sqrt(u.^2 + v.^2);
            maxDisplacement = max(deformation);
            if (maxDisplacement == 0)
                disp('Zero deformation encountered');
            end
            % Length of each spring
            [xTemplate,yTemplate] = obj.springTemplate();
            xy = obj.myNodeLocations;
            xy = xy + uv;
            C = obj.myConnectivity;
            startPts = xy(:,C(1,:));
            endPts = xy(:,C(2,:));
            L = sqrt((endPts(1,:)-startPts(1,:)).^2 + (endPts(2,:)-startPts(2,:)).^2);
            theta = atan2(endPts(2,:)-startPts(2,:),endPts(1,:)-startPts(1,:));
            ct = cos(theta);
            st = sin(theta);
            for j = 1:obj.myNumNodes % for all nodes
                x0 = xy(1,j); y0 = xy(2,j);
                h = plot(x0,y0,'ro');set(h,'LineWidth',2.0);
            end
            for s = 1:obj.myNumSpringElements
                xStretch = xTemplate*L(s);
                yStretch = yTemplate;
                xRotate = xStretch*ct(s) - yStretch*st(s);
                yRotate = xStretch*st(s) + yStretch*ct(s);
                xSpring = xRotate + xy(1,obj.myConnectivity(1,s));
                ySpring = yRotate + xy(2,obj.myConnectivity(1,s));
                plot(xSpring,ySpring,'r','linewidth',1.0);hold on;
            end
            title(['Max deformation = ' num2str(maxDisplacement)]);
            axis equal;axis tight;
        end
        
    end
end
